"""REST API video device interface

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""
import os
from typing import List
import logging
import re
import tempfile
import subprocess
import shlex

from fastapi import APIRouter
from fastapi import Depends
from fastapi.responses import JSONResponse
from fastapi import HTTPException
from pydantic import BaseModel
from pydantic import Field

import qamlib
import quarklib

log = logging.getLogger("pion_dut")


router = APIRouter(prefix="/pion")


class DeviceList(BaseModel):
    """List of video devices in the camera"""

    video: List[str] = Field(
        description="generic video devices",
        example=[
            "/dev/video5",
            "/dev/video4",
            "/dev/video3",
            "/dev/video2",
            "/dev/video1",
            "/dev/video0",
        ],
    )
    """: generic video devices"""
    qtec_video: List[str] = Field(
        description="qtec video capture devices", example=["/dev/qtec/video0"]
    )
    """: qtec video capture devices"""
    qtec: List[str] = Field(
        description="qtec video devices",
        example=[
            "/dev/qtec/xform_dist_readback0",
            "/dev/qtec/xform_gain0",
            "/dev/qtec/xform_dist0",
            "/dev/qtec/video0",
            "/dev/qtec/xform_gain_readback0",
        ],
    )
    """: qtec video devices"""


def get_video_device_nrs(path: str = "/dev", name: str = "video") -> List[int]:
    """List video devices numbers

    Filter based on specified name pattern

    :param path: path to look for video devices
    :param name: name pattern used to filter video devices
    :return: list of video device numbers
    """
    devs = os.listdir(path)
    vid_indices = [int(dev[-1]) for dev in devs if dev.startswith(name)]
    vid_indices = sorted(vid_indices)
    return vid_indices


def list_video_devices(path: str = "/dev", name: str = "video") -> List[str]:
    """List video devices with specified name pattern

    :param path: path to look for video devices
    :param name: name pattern used to filter video devices
    :return: list of video device names (full path)
    """
    devs = os.listdir(path)
    devices = [f"{path}/{dev}" for dev in devs if dev.startswith(name)]
    return devices


def list_qtec_video_devices() -> List[str]:
    """List qtec video capture devices"""
    return list_video_devices("/dev/qtec", name="video")


def list_qtec_devices() -> List[str]:
    """List qtec video devices"""
    return list_video_devices("/dev/qtec", name="")


@router.get(
    "/list_devices",
    response_model=DeviceList,
    summary="List video devices",
    response_description="Lists of different video devices",
)
def list_devices() -> JSONResponse:
    """List video devices (both generic and qtec)"""
    return DeviceList(
        video=list_video_devices(),
        qtec_video=list_qtec_video_devices(),
        qtec=list_qtec_devices(),
    )

cams = {}

@router.get("/open_cams")
def open_cams() -> JSONResponse:
    """Open camera devices.

    Opens all video devices under ``/dev/qtec``
    """
    global cams
    for dev in os.listdir("/dev/qtec"):
        if dev.startswith("video"):
            try:
                cams[dev] = qamlib.Camera(os.path.join("/dev/qtec", dev))
            except Exception as exc:
                log.exception(exc)
                raise HTTPException(status_code=500, detail=f"Failed to open {dev}")

    return {"num_devices": len(cams)}

@router.get("/start_stream")
def start_stream() -> None:
    """Start camera stream.

    Starts streaming for all open camera devices. Use ``open_cams`` to open
    devices.
    """

    for dev in cams:
        try:
            cams[dev].start()
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed starting stream for {dev}")

@router.get("/stop_stream")
def stop_stream() -> None:
    """Start camera stream.

    Stops streaming for all open camera devices.
    """

    for dev in cams:
        try:
            cams[dev].stop()
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed stopping stream for {dev}")

@router.get("/set_exposure")
def set_exposure(exp: int) -> None:
    """Set the exposure time

    Sets the exposure for all open camera devices.
    """

    for dev in cams:
        try:
            cams[dev].set_control("exposure_time_absolute", exp)
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed setting exposure for {dev}")

@router.get("/set_self_timed")
def set_self_timed() -> None:
    """Set self timed

    Sets the trigger mode to ``self_timed`` for all open cameras.
    """

    for dev in cams:
        try:
            cams[dev].set_control("trigger_mode", 0)
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed setting external trigger for {dev}")

@router.get("/set_external_trigger")
def set_external_trigger() -> None:
    """Set external trigger

    Sets the trigger mode to ``external_trigger`` for all open cameras.
    """

    for dev in cams:
        try:
            cams[dev].set_control("trigger_mode", 1)
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed setting external trigger for {dev}")

@router.get("/set_external_exposure")
def set_external_exposure() -> None:
    """Set external exposure

    Sets the trigger mode to ``external_exposure`` for all open cameras.
    """

    for dev in cams:
        try:
            cams[dev].set_control("trigger_mode", 2)
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed setting external exposure for {dev}")

@router.get("/get_latest_sequence_num")
def get_latest_sequence_num() -> JSONResponse:
    """Gets the latest frame sequence number

    For each open camera get the last frame sequence number. The cameras need
    to be streaming.

    This runs until it either times out getting the next frame, which is
    interpreted as the last frame. Or it gets another exception.
    """
    seqs = {}

    for dev in cams:
        try:
            seq = -1
            while True:
                try:
                    logging.warning("Getting frame")
                    meta, _ = cams[dev].get_frame(timeout=1)
                    logging.warning(f"got frame {meta.sequence}")
                    seq = meta.sequence
                except qamlib.TimeoutException:
                    seqs[dev] = seq
                    break
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed getting sequence number for {dev}")

    return seqs

black_level_regex = re.compile(r"(BLACK_LEVEL:) (\dx[0-9a-f]+)$")
def black_level_calibrate_and_save(cam):
    i2c_bus = cam.get_control("head_i2c_bus")
    i2c_addr = cam.get_control("head_i2c_address")


    with tempfile.NamedTemporaryFile() as temp:
        cmd = shlex.split(f"qtec-eeprom-read /dev/i2c-{i2c_bus} {i2c_addr} {temp.name}")
        print(cmd)
        ret = subprocess.run(cmd)

        if ret.returncode != 0:
            return False

        data = temp.read().decode()

    print(f"I2C Bus: {i2c_bus}, I2C Addr: {i2c_addr}")
    black_level = -1
    for line in data.split("\n"):
        m = black_level_regex.match(line)
        if m is None:
            continue

        black_level = int(m.group(2), 16)
        break
    else:
        print("Error reading EEPROM")
        print(data)
        return False

    quarklib.isp.calibrate_black_level(cam)

    new_level = cam.get_control("black_level")
    print(f"Got black level: {new_level}")
    if new_level == black_level:
        return True


    new = ""
    for line in data.split("\n"):
        new += black_level_regex.sub(rf"\1 {hex(new_level)}", line) + "\n"


    with tempfile.NamedTemporaryFile(delete=False) as temp:
        new_data = bytes(new, encoding="utf-8")
        temp.write(new_data)
        temp.close()

        with open(temp.name) as fd:
            print(fd.read())

        cmd = shlex.split(f"qtec-eeprom-write /dev/i2c-{i2c_bus} {i2c_addr} {temp.name}")
        ret = subprocess.run(cmd)

        os.remove(temp.name)

    return new_level


@router.get("/calibrate_black_level")
def calibrate_black_level() -> JSONResponse:
    """Calibrates black level for all cameras."""

    levels = {}
    for dev in cams:
        try:
            levels[dev] = black_level_calibrate_and_save(cams[dev])
        except Exception as exc:
            log.exception(exc)
            raise HTTPException(status_code=500, detail=f"Failed calibrating black level for {dev}")

    return levels

