"""Quark API Backend for Qtec cameras

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""
import logging
import os.path
import sys
import threading
from typing import List
from typing import Optional

import uvicorn
from fastapi import FastAPI
from fastapi import Request
from fastapi.responses import HTMLResponse
from fastapi.responses import JSONResponse

from .api import router


description = """
API for use in the camera production by the Pion test system.
"""

tags_meta = [
    {
        "name": "Pion API",
        "description": "Miscallenous API calls used for Pion",
    },
]

app = FastAPI(
    title="Pion Server",
    description=description,
    summary="API for the Pion for production of cameras.",
    version="0.1.0",
    license_info={
        "name": "LGPL2.1",
        "url": "https://opensource.org/licenses/LGPL-2.1",
    },
    openapi_tags=tags_meta,
)

app.openapi_version = "3.0.3"

# include routers
app.include_router(router)


@app.on_event("shutdown")
def shutdown_event():
    """Shutdown event"""
    pass


def handle_args(cli_args: Optional[List[str]] = None) -> List[str]:
    """Command-line arguments

    `cli_args` makes it possible to call functions command-line-style
    from other Python code without touching sys.argv.
    """
    if cli_args is None:
        return sys.argv[1:]
    else:
        return cli_args


def generate_index_html():
    html_content = """
    <html>
        <head>
            <title>Pion API</title>
        </head>
        <body>
            <h1>API for interacting with Qtechnology A/S cameras.</h1>
            <ul>
                <li><a href="docs">Live API documentation</a></li>
            </ul>
        </body>
    </html>
    """
    return HTMLResponse(content=html_content, status_code=200)


@app.get("/", response_class=HTMLResponse)
def home() -> HTMLResponse:
    """index.html"""
    return generate_index_html()


def main(cli_args: Optional[List[str]] = None):
    """Fast API App

    `cli_args` makes it possible to call this function command-line-style
    from other Python code without touching sys.argv.
    """
    #log = logger_conf("main", c_level=logging.DEBUG)

    # app.logger.warning("Test")

    DEBUG_MODE = False

    # start server thread
    #log.debug("Starting Uvicorn server thread")
    uvicorn.run(
        "pion_dut.main:app",
        host="0.0.0.0",  # noqa: S104
        port=5000,
        log_level="debug" if DEBUG_MODE else "info",
        reload=DEBUG_MODE,
        # reload_dirs=[],
    )

    #log.debug(f"Current number of threads: {threading.active_count()}")

if __name__ == "__main__":
    main()
